#include <ppl.hh>
#include <gmpxx.h>
#include <cstdlib>
#include <fstream>
#include <chrono>

using namespace std::chrono;
using namespace Parma_Polyhedra_Library;
using namespace Parma_Polyhedra_Library::IO_Operators;
using namespace std;

void read_signal_file(char *fName, vector<int> &value_list){
    ifstream infile(fName);
    string line;
    int lineNum=0;
    while(std::getline(infile,line)){
        string s=line;
        if(s.find(",")==string::npos){
            continue;
        }
        int pos = s.find(",");
        int listart = atoi(s.substr(0,pos).c_str());
        int liend   = atoi(s.substr(s.find(",")+1).c_str());
        value_list.push_back(liend);
    }
    infile.close();
}

int constraint_complement(Constraint &c_in, Constraint &cs_out){
    // cout << "Constraint c_in: " << c_in << endl;
    if (c_in.is_equality()){
      Linear_Expression e;
      for (dimension_type i = c_in.space_dimension(); i-- > 0; )
        e += c_in.coefficient(Variable(i)) * Variable(i);
      e += c_in.inhomogeneous_term();
      // cout << "Constraint c_in is not an inequality." << endl;
      return 0;
    } else {
      Linear_Expression e;
      for (dimension_type i = c_in.space_dimension(); i-- > 0; )
        e += c_in.coefficient(Variable(i)) * Variable(i);
      e += c_in.inhomogeneous_term();
      cs_out = c_in.is_strict_inequality() ? (e <= 0) : (e < 0);
      // cout << "Complement c_out: " << cs_out << endl;
      return 1;
    }

}


Pointset_Powerset<NNC_Polyhedron> negate_polyhedron(NNC_Polyhedron &nnc_in, Constraint_System &param_domain){
    Pointset_Powerset<NNC_Polyhedron> pset_res;
    Constraint_System cs = nnc_in.minimized_constraints();
    int suc_res = 0;

    for (Constraint_System_const_iterator i = cs.begin(),
        cs_end = cs.end(); i != cs_end; ++i){
        Constraint ctemp(*i), ctemp_neg;
        int res = constraint_complement(ctemp, ctemp_neg);
        if(res){
            Constraint_System pd_temp = param_domain;
            pd_temp.insert(ctemp_neg);
            if(suc_res == 0){
                pset_res = Pointset_Powerset<NNC_Polyhedron>(NNC_Polyhedron(pd_temp));
                suc_res = 1;
            }else{
                pset_res.add_disjunct(NNC_Polyhedron(pd_temp));
            }
        }else{
            Linear_Expression e;
            for (dimension_type i = ctemp.space_dimension(); i-- > 0; )
                e += ctemp.coefficient(Variable(i)) * Variable(i);
            e += ctemp.inhomogeneous_term();
            Constraint_System pd_temp_1 = param_domain;
            pd_temp_1.insert(e>0);
            Constraint_System pd_temp_2 = param_domain;
            pd_temp_2.insert(e<0);
            if(suc_res == 0){
                pset_res = Pointset_Powerset<NNC_Polyhedron>(NNC_Polyhedron(pd_temp_1));
                pset_res.add_disjunct(NNC_Polyhedron(pd_temp_2));
                suc_res = 1;
            }else{
                pset_res.add_disjunct(NNC_Polyhedron(pd_temp_1));
                pset_res.add_disjunct(NNC_Polyhedron(pd_temp_2));
            }
        }
    }
    return pset_res;
}

Pointset_Powerset<NNC_Polyhedron> negate_nnc_powerset(Pointset_Powerset<NNC_Polyhedron> ps,
    Constraint_System param_domain){

    Pointset_Powerset<NNC_Polyhedron> ps_res;
    int suc_res = 0;
    int count = 0;
    for(auto nncp : ps){
        Constraint_System cs = nncp.pointset().minimized_constraints();
        NNC_Polyhedron nnc_temp(cs);
        Pointset_Powerset<NNC_Polyhedron> neg_nnc = negate_polyhedron(nnc_temp, param_domain);
        if(suc_res == 0){
            ps_res = neg_nnc;
        }else{
            ps_res.intersection_assign(neg_nnc);
        }
        suc_res = 1;
        count++;
    }
    return ps_res;
}


int main(int argc, char** argv){
	vector<int> signal_vec;
    read_signal_file(argv[1], signal_vec);

    auto start = high_resolution_clock::now();
    
    int max_integral = 0;
    for(int i=0; i<signal_vec.size(); i++){
    	int elem = signal_vec[i];
        max_integral += elem;
    }

    cout<<"integral:"<<max_integral<<endl;

    Variable p1(0),p2(1);
    // Create validity domain
    Constraint p1c1(0<=p1),p1c2(p1<=signal_vec.size());
    Constraint p2c1(0<=p2),p2c2(p2<=max_integral);
    Constraint_System param_domain;
    param_domain.insert(p1c1);
    param_domain.insert(p1c2);
    param_domain.insert(p2c1);
    param_domain.insert(p2c2);
    Pointset_Powerset<NNC_Polyhedron> vdomain;
    int acc_integral=0;
    for(int i=0; i<signal_vec.size(); i++){
        Constraint_System cs = param_domain;
        Constraint c1(i<=p1),c2(p1<(i+1));
        Constraint c3((acc_integral+signal_vec[i]*(p1-i))>=(max_integral-p2));
        cs.insert(c1);
        cs.insert(c2);
        cs.insert(c3);
        NNC_Polyhedron nncp(cs);
        if(i == 0){
            vdomain = Pointset_Powerset<NNC_Polyhedron>(nncp);
        }else{
            vdomain.add_disjunct(nncp);    
        }
        acc_integral += signal_vec[i];
    }
    cout<<"------------"<<endl;
    cout<<vdomain<<endl;
    cout<<"-------------"<<endl;

    Pointset_Powerset<NNC_Polyhedron> ppold = vdomain;
    int num_negations = atoi(argv[2]);
    for(int i=0; i<num_negations; i++){
        Pointset_Powerset<NNC_Polyhedron> ppnew = negate_nnc_powerset(ppold, param_domain);
        ppold = ppnew;
    }
    cout<<"------------"<<endl;
    cout<<ppold<<endl;
    cout<<"-------------"<<endl;

    auto stop = high_resolution_clock::now();
    auto duration = duration_cast<milliseconds>(stop-start);
    cout<<"time: "<<duration.count()<<" ms"<<endl;

	return 0;	
}